from dataclasses import dataclass
import uuid
from typing import Union, Sequence, Optional

import pytest

import weaviate


UUID = Union[str, uuid.UUID]


def has_batch_errors(results: dict) -> bool:
    """
    Check batch results for errors.

    Parameters
    ----------
    results : dict
        The Weaviate batch creation return value.
    """

    if results is not None:
        for result in results:
            if "result" in result and "errors" in result["result"]:
                if "error" in result["result"]["errors"]:
                    return True
    return False


@dataclass
class MockNumpyTorch:
    array: list

    def squeeze(self) -> "MockNumpyTorch":
        return self

    def tolist(self) -> list:
        return self.array


@dataclass
class MockTensorFlow:
    array: list

    def numpy(self) -> "MockNumpyTorch":
        return MockNumpyTorch(self.array)


@pytest.fixture(scope="function")
def client():
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    client.schema.create_class(
        {
            "class": "Test",
            "properties": [{"name": "test", "dataType": ["Test"]}],
            "vectorizer": "none",
        }
    )
    yield client
    client.schema.delete_all()


@pytest.mark.parametrize(
    "vector",
    [None, [1, 2, 3], MockNumpyTorch([1, 2, 3]), MockTensorFlow([1, 2, 3])],
)
@pytest.mark.parametrize("uuid", [None, uuid.uuid4(), str(uuid.uuid4()), uuid.uuid4().hex])
def test_add_data_object(client: weaviate.Client, uuid: Optional[UUID], vector: Optional[Sequence]):
    """Test the `add_data_object` method"""
    client.batch.add_data_object(
        data_object={},
        class_name="Test",
        uuid=uuid,
        vector=vector,
    )
    response = client.batch.create_objects()
    assert has_batch_errors(response) is False, str(response)


@pytest.mark.parametrize("from_object_uuid", [uuid.uuid4(), str(uuid.uuid4()), uuid.uuid4().hex])
@pytest.mark.parametrize("to_object_uuid", [uuid.uuid4().hex, uuid.uuid4(), str(uuid.uuid4())])
@pytest.mark.parametrize("to_object_class_name", [None, "Test"])
def test_add_reference(
    client: weaviate.Client,
    from_object_uuid: UUID,
    to_object_uuid: UUID,
    to_object_class_name: Optional[str],
):
    """Test the `add_reference` method"""

    # create the 2 objects first
    client.data_object.create(
        data_object={},
        class_name="Test",
        uuid=from_object_uuid,
    )
    client.data_object.create(
        data_object={},
        class_name="Test",
        uuid=to_object_uuid,
    )

    client.batch.add_reference(
        from_object_uuid=from_object_uuid,
        from_object_class_name="Test",
        from_property_name="test",
        to_object_uuid=to_object_uuid,
        to_object_class_name=to_object_class_name,
    )

    response = client.batch.create_references()
    assert has_batch_errors(response) is False, str(response)



from typing import Optional

import pytest

import weaviate


@pytest.fixture(scope="module")
def client():
    client = weaviate.Client("http://localhost:8080")
    yield client
    client.schema.delete_all()


@pytest.mark.parametrize("replicationFactor", [None, 1])
def test_create_class_with_implicit_and_explicit_replication_factor(
    client: weaviate.Client, replicationFactor: Optional[int]
):
    single_class = {
        "class": "Barbecue",
        "description": "Barbecue or BBQ where meat and vegetables get grilled",
        "properties": [
            {
                "dataType": ["string"],
                "description": "how hot is the BBQ in C",
                "name": "heat",
            },
        ],
    }
    if replicationFactor is None:
        expected_factor = 1
    else:
        expected_factor = replicationFactor
        single_class["replicationConfig"] = {
            "factor": replicationFactor,
        }

    client.schema.create_class(single_class)
    created_class = client.schema.get("Barbecue")
    assert created_class["class"] == "Barbecue"
    assert created_class["replicationConfig"]["factor"] == expected_factor

    client.schema.delete_class("Barbecue")


@pytest.mark.parametrize("data_type", ["uuid", "uuid[]"])
def test_uuid_datatype(client, data_type):
    single_class = {"class": "UuidTest", "properties": [{"dataType": [data_type], "name": "heat"}]}

    client.schema.create_class(single_class)
    created_class = client.schema.get("uuidTest")
    assert created_class["class"] == "UuidTest"

    client.schema.delete_class("UuidTest")


@pytest.mark.parametrize("tokenization", ["word", "whitespace", "lowercase", "field"])
def test_tokenization(client, tokenization):
    single_class = {
        "class": "TokenTest",
        "properties": [{"dataType": ["text"], "name": "heat", "tokenization": tokenization}],
    }
    client.schema.create_class(single_class)
    created_class = client.schema.get("TokenTest")
    assert created_class["class"] == "TokenTest"

    client.schema.delete_class("TokenTest")


def test_class_exists(client: weaviate.Client):
    single_class = {"class": "Exists"}

    client.schema.create_class(single_class)
    assert client.schema.exists("Exists") is True
    assert client.schema.exists("DoesNotExists") is False

    client.schema.delete_class("Exists")
    assert client.schema.exists("Exists") is False


def test_schema_keys(client: weaviate.Client):
    single_class = {
        "class": "Author",
        "properties": [
            {
                "indexFilterable": False,
                "indexSearchable": False,
                "dataType": ["text"],
                "name": "name",
            }
        ],
    }
    client.schema.create_class(single_class)
    assert client.schema.exists("Author")



from dataclasses import dataclass, field
from typing import Optional


@dataclass
class ConnectionConfig:
    session_pool_connections: int = 20
    session_pool_maxsize: int = 20

    def __post_init__(self) -> None:
        if not isinstance(self.session_pool_connections, int):
            raise TypeError(
                f"session_pool_connections must be {int}, received {type(self.session_pool_connections)}"
            )
        if not isinstance(self.session_pool_maxsize, int):
            raise TypeError(
                f"session_pool_maxsize must be {int}, received {type(self.session_pool_maxsize)}"
            )


@dataclass
class Config:
    grpc_port_experimental: Optional[int] = None
    connection_config: ConnectionConfig = field(default_factory=ConnectionConfig)

    def __post_init__(self) -> None:
        if self.grpc_port_experimental is not None and not isinstance(
            self.grpc_port_experimental, int
        ):
            raise TypeError(
                f"grpc_port_experimental must be {int}, received {type(self.grpc_port_experimental)}"
            )



from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AdditionalProperties(_message.Message):
    __slots__ = ["certainty", "creationTimeUnix", "distance", "explainScore", "lastUpdateTimeUnix", "score", "uuid", "vector"]
    CERTAINTY_FIELD_NUMBER: _ClassVar[int]
    CREATIONTIMEUNIX_FIELD_NUMBER: _ClassVar[int]
    DISTANCE_FIELD_NUMBER: _ClassVar[int]
    EXPLAINSCORE_FIELD_NUMBER: _ClassVar[int]
    LASTUPDATETIMEUNIX_FIELD_NUMBER: _ClassVar[int]
    SCORE_FIELD_NUMBER: _ClassVar[int]
    UUID_FIELD_NUMBER: _ClassVar[int]
    VECTOR_FIELD_NUMBER: _ClassVar[int]
    certainty: bool
    creationTimeUnix: bool
    distance: bool
    explainScore: bool
    lastUpdateTimeUnix: bool
    score: bool
    uuid: bool
    vector: bool
    def __init__(self, uuid: bool = ..., vector: bool = ..., creationTimeUnix: bool = ..., lastUpdateTimeUnix: bool = ..., distance: bool = ..., certainty: bool = ..., score: bool = ..., explainScore: bool = ...) -> None: ...

class BM25SearchParams(_message.Message):
    __slots__ = ["properties", "query"]
    PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    QUERY_FIELD_NUMBER: _ClassVar[int]
    properties: _containers.RepeatedScalarFieldContainer[str]
    query: str
    def __init__(self, query: _Optional[str] = ..., properties: _Optional[_Iterable[str]] = ...) -> None: ...

class HybridSearchParams(_message.Message):
    __slots__ = ["alpha", "properties", "query", "vector"]
    ALPHA_FIELD_NUMBER: _ClassVar[int]
    PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    QUERY_FIELD_NUMBER: _ClassVar[int]
    VECTOR_FIELD_NUMBER: _ClassVar[int]
    alpha: float
    properties: _containers.RepeatedScalarFieldContainer[str]
    query: str
    vector: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, query: _Optional[str] = ..., properties: _Optional[_Iterable[str]] = ..., vector: _Optional[_Iterable[float]] = ..., alpha: _Optional[float] = ...) -> None: ...

class NearObjectParams(_message.Message):
    __slots__ = ["certainty", "distance", "id"]
    CERTAINTY_FIELD_NUMBER: _ClassVar[int]
    DISTANCE_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    certainty: float
    distance: float
    id: str
    def __init__(self, id: _Optional[str] = ..., certainty: _Optional[float] = ..., distance: _Optional[float] = ...) -> None: ...

class NearVectorParams(_message.Message):
    __slots__ = ["certainty", "distance", "vector"]
    CERTAINTY_FIELD_NUMBER: _ClassVar[int]
    DISTANCE_FIELD_NUMBER: _ClassVar[int]
    VECTOR_FIELD_NUMBER: _ClassVar[int]
    certainty: float
    distance: float
    vector: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, vector: _Optional[_Iterable[float]] = ..., certainty: _Optional[float] = ..., distance: _Optional[float] = ...) -> None: ...

class Properties(_message.Message):
    __slots__ = ["non_ref_properties", "ref_properties"]
    NON_REF_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    REF_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    non_ref_properties: _containers.RepeatedScalarFieldContainer[str]
    ref_properties: _containers.RepeatedCompositeFieldContainer[RefProperties]
    def __init__(self, non_ref_properties: _Optional[_Iterable[str]] = ..., ref_properties: _Optional[_Iterable[_Union[RefProperties, _Mapping]]] = ...) -> None: ...

class RefProperties(_message.Message):
    __slots__ = ["linked_class", "linked_properties", "reference_property"]
    LINKED_CLASS_FIELD_NUMBER: _ClassVar[int]
    LINKED_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    REFERENCE_PROPERTY_FIELD_NUMBER: _ClassVar[int]
    linked_class: str
    linked_properties: Properties
    reference_property: str
    def __init__(self, linked_class: _Optional[str] = ..., reference_property: _Optional[str] = ..., linked_properties: _Optional[_Union[Properties, _Mapping]] = ...) -> None: ...

class ResultAdditionalProps(_message.Message):
    __slots__ = ["certainty", "certainty_present", "creation_time_unix", "creation_time_unix_present", "distance", "distance_present", "explain_score", "explain_score_present", "id", "last_update_time_unix", "last_update_time_unix_present", "score", "score_present", "vector"]
    CERTAINTY_FIELD_NUMBER: _ClassVar[int]
    CERTAINTY_PRESENT_FIELD_NUMBER: _ClassVar[int]
    CREATION_TIME_UNIX_FIELD_NUMBER: _ClassVar[int]
    CREATION_TIME_UNIX_PRESENT_FIELD_NUMBER: _ClassVar[int]
    DISTANCE_FIELD_NUMBER: _ClassVar[int]
    DISTANCE_PRESENT_FIELD_NUMBER: _ClassVar[int]
    EXPLAIN_SCORE_FIELD_NUMBER: _ClassVar[int]
    EXPLAIN_SCORE_PRESENT_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    LAST_UPDATE_TIME_UNIX_FIELD_NUMBER: _ClassVar[int]
    LAST_UPDATE_TIME_UNIX_PRESENT_FIELD_NUMBER: _ClassVar[int]
    SCORE_FIELD_NUMBER: _ClassVar[int]
    SCORE_PRESENT_FIELD_NUMBER: _ClassVar[int]
    VECTOR_FIELD_NUMBER: _ClassVar[int]
    certainty: float
    certainty_present: bool
    creation_time_unix: int
    creation_time_unix_present: bool
    distance: float
    distance_present: bool
    explain_score: str
    explain_score_present: bool
    id: str
    last_update_time_unix: int
    last_update_time_unix_present: bool
    score: float
    score_present: bool
    vector: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, id: _Optional[str] = ..., vector: _Optional[_Iterable[float]] = ..., creation_time_unix: _Optional[int] = ..., creation_time_unix_present: bool = ..., last_update_time_unix: _Optional[int] = ..., last_update_time_unix_present: bool = ..., distance: _Optional[float] = ..., distance_present: bool = ..., certainty: _Optional[float] = ..., certainty_present: bool = ..., score: _Optional[float] = ..., score_present: bool = ..., explain_score: _Optional[str] = ..., explain_score_present: bool = ...) -> None: ...

class ResultProperties(_message.Message):
    __slots__ = ["class_name", "non_ref_properties", "ref_props"]
    CLASS_NAME_FIELD_NUMBER: _ClassVar[int]
    NON_REF_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    REF_PROPS_FIELD_NUMBER: _ClassVar[int]
    class_name: str
    non_ref_properties: _struct_pb2.Struct
    ref_props: _containers.RepeatedCompositeFieldContainer[ReturnRefProperties]
    def __init__(self, non_ref_properties: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ..., ref_props: _Optional[_Iterable[_Union[ReturnRefProperties, _Mapping]]] = ..., class_name: _Optional[str] = ...) -> None: ...

class ReturnRefProperties(_message.Message):
    __slots__ = ["prop_name", "properties"]
    PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    PROP_NAME_FIELD_NUMBER: _ClassVar[int]
    prop_name: str
    properties: _containers.RepeatedCompositeFieldContainer[ResultProperties]
    def __init__(self, properties: _Optional[_Iterable[_Union[ResultProperties, _Mapping]]] = ..., prop_name: _Optional[str] = ...) -> None: ...

class SearchReply(_message.Message):
    __slots__ = ["results", "took"]
    RESULTS_FIELD_NUMBER: _ClassVar[int]
    TOOK_FIELD_NUMBER: _ClassVar[int]
    results: _containers.RepeatedCompositeFieldContainer[SearchResult]
    took: float
    def __init__(self, results: _Optional[_Iterable[_Union[SearchResult, _Mapping]]] = ..., took: _Optional[float] = ...) -> None: ...

class SearchRequest(_message.Message):
    __slots__ = ["additional_properties", "bm25_search", "class_name", "hybrid_search", "limit", "near_object", "near_vector", "properties"]
    ADDITIONAL_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    BM25_SEARCH_FIELD_NUMBER: _ClassVar[int]
    CLASS_NAME_FIELD_NUMBER: _ClassVar[int]
    HYBRID_SEARCH_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    NEAR_OBJECT_FIELD_NUMBER: _ClassVar[int]
    NEAR_VECTOR_FIELD_NUMBER: _ClassVar[int]
    PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    additional_properties: AdditionalProperties
    bm25_search: BM25SearchParams
    class_name: str
    hybrid_search: HybridSearchParams
    limit: int
    near_object: NearObjectParams
    near_vector: NearVectorParams
    properties: Properties
    def __init__(self, class_name: _Optional[str] = ..., limit: _Optional[int] = ..., additional_properties: _Optional[_Union[AdditionalProperties, _Mapping]] = ..., near_vector: _Optional[_Union[NearVectorParams, _Mapping]] = ..., near_object: _Optional[_Union[NearObjectParams, _Mapping]] = ..., properties: _Optional[_Union[Properties, _Mapping]] = ..., hybrid_search: _Optional[_Union[HybridSearchParams, _Mapping]] = ..., bm25_search: _Optional[_Union[BM25SearchParams, _Mapping]] = ...) -> None: ...

class SearchResult(_message.Message):
    __slots__ = ["additional_properties", "properties"]
    ADDITIONAL_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    additional_properties: ResultAdditionalProps
    properties: ResultProperties
    def __init__(self, properties: _Optional[_Union[ResultProperties, _Mapping]] = ..., additional_properties: _Optional[_Union[ResultAdditionalProps, _Mapping]] = ...) -> None: ...



import unittest
from weaviate.schema.validate_schema import validate_schema, check_class, check_property
from weaviate.exceptions import SchemaValidationException
from test.util import check_error_message


valid_schema_with_all_properties = {
    "classes": [
        {
            "class": "Category",
            "description": "Category an article is a type off",
            "moduleConfig": {"text2vec-contextionary": {"vectorizeClassName": False}},
            "properties": [
                {
                    "dataType": ["string"],
                    "description": "category name",
                    "indexInverted": True,
                    "moduleConfig": {"text2vec-contextionary": {"vectorizePropertyName": False}},
                    "name": "name",
                }
            ],
            "vectorIndexType": "hnsw",
            "vectorizer": "none",
            "replicationConfig": {
                "factor": 1,
            },
        },
        {
            "class": "Publication",
            "description": "A publication with an online source",
            "moduleConfig": {"text2vec-contextionary": {"vectorizeClassName": False}},
            "properties": [
                {"dataType": ["string"], "description": "Name of the publication", "name": "name"},
                {
                    "dataType": ["geoCoordinates"],
                    "description": "Geo location of the HQ",
                    "name": "headquartersGeoLocation",
                },
                {
                    "dataType": ["Article"],
                    "description": "The articles this publication has",
                    "name": "hasArticles",
                },
                {
                    "dataType": ["Article"],
                    "description": "Articles this author wrote",
                    "name": "wroteArticles",
                },
            ],
            "vectorIndexType": "hnsw",
            "vectorizer": "none",
            "replicationConfig": {
                "factor": 1,
            },
        },
        {
            "class": "Author",
            "description": "Normalised types",
            "moduleConfig": {"text2vec-contextionary": {"vectorizeClassName": True}},
            "properties": [
                {"dataType": ["string"], "description": "Name of the author", "name": "name"},
                {
                    "dataType": ["Publication"],
                    "description": "The publication this author writes for",
                    "name": "writesFor",
                },
            ],
            "vectorIndexType": "hnsw",
            "vectorizer": "none",
            "replicationConfig": {
                "factor": 1,
            },
        },
        {
            "class": "Article",
            "description": "Normalised types",
            "moduleConfig": {"text2vec-contextionary": {"vectorizeClassName": False}},
            "properties": [
                {
                    "dataType": ["string"],
                    "description": "title of the article",
                    "indexInverted": True,
                    "moduleConfig": {"text2vec-contextionary": {"vectorizePropertyName": False}},
                    "name": "title",
                },
                {
                    "dataType": ["string"],
                    "description": "url of the article",
                    "indexInverted": False,
                    "moduleConfig": {"text2vec-contextionary": {"vectorizePropertyName": False}},
                    "name": "url",
                },
                {
                    "dataType": ["text"],
                    "description": "summary of the article",
                    "indexInverted": True,
                    "moduleConfig": {"text2vec-contextionary": {"vectorizePropertyName": False}},
                    "name": "summary",
                },
                {
                    "dataType": ["date"],
                    "description": "date of publication of the article",
                    "name": "publicationDate",
                },
                {"dataType": ["int"], "description": "Words in this article", "name": "wordCount"},
                {
                    "dataType": ["Author", "Publication"],
                    "description": "authors this article has",
                    "name": "hasAuthors",
                },
                {
                    "dataType": ["Publication"],
                    "description": "publication this article is in",
                    "name": "inPublication",
                },
                {
                    "dataType": ["Category"],
                    "description": "category this article is of",
                    "name": "ofCategory",
                },
                {
                    "dataType": ["boolean"],
                    "description": "whether the article is currently accessible through the url",
                    "name": "isAccessible",
                },
            ],
            "vectorIndexType": "hnsw",
            "vectorizer": "none",
            "replicationConfig": {
                "factor": 1,
            },
        },
    ]
}


class TestSchemaValidation(unittest.TestCase):
    def test_validate_schema(self):
        """
        Test `validate_schema` function.
        """

        # invalid calls
        classess_error_message = (
            'Each schema has to have "classes" '
            "in the first level of the JSON format file/parameter/object"
        )
        class_key_error_message = '"class" key is missing in class definition.'

        invalid_schema = {}
        with self.assertRaises(SchemaValidationException) as error:
            validate_schema(invalid_schema)
        check_error_message(self, error, classess_error_message)

        invalid_schema = {"classes": "my_class"}
        with self.assertRaises(SchemaValidationException) as error:
            validate_schema(invalid_schema)
        check_error_message(self, error, f'"classes" is type {str} but should be {list}.')

        invalid_schema = {"things": {"classes": []}}
        with self.assertRaises(SchemaValidationException) as error:
            validate_schema(invalid_schema)
        check_error_message(self, error, classess_error_message)

        invalid_schema = {"classes": ["my_class"]}
        with self.assertRaises(SchemaValidationException) as error:
            validate_schema(invalid_schema)
        check_error_message(self, error, f'"class" is type {str} but should be {dict}.')

        # test the call of the `check_class` function inside `validate_schema`
        invalid_schema = {"classes": [{"my_class": []}]}
        with self.assertRaises(SchemaValidationException) as error:
            validate_schema(invalid_schema)
        check_error_message(self, error, class_key_error_message)

        # valid calls
        valid_schema = {"classes": []}
        self.assertIsNone(validate_schema(valid_schema))
        valid_schema = {"classes": [], "author": "Unit Test"}
        self.assertIsNone(validate_schema(valid_schema))
        self.assertIsNone(validate_schema(valid_schema_with_all_properties))

    def test_check_class(self):
        """
        Test `check_class` function.
        """

        # Valid maximal schema
        max_valid = {
            "class": "Boat",
            "description": "boat swiming on the water",
            "properties": [],
            "vectorIndexType": "hnsw",
            "vectorIndexConfig": {},
            "moduleConfig": {},
            "vectorizer": "text2vec-contextionary",
            "replicationConfig": {
                "factor": 1,
            },
        }
        check_class(max_valid)
        # minimal must contain class key as string
        check_class({"class": "Car"})

        # invalid calls
        class_key_error_message = '"class" key is missing in class definition.'
        unknown_key_error_message = lambda key: f'"{key}" is not a known class definition key.'
        key_type_error_messsage = lambda key, value, exp_type: (
            f'"{key}" is type {type(value)} ' f"but should be {exp_type}."
        )

        with self.assertRaises(SchemaValidationException) as error:
            check_class({"invalid_key": "value"})
        check_error_message(self, error, class_key_error_message)

        with self.assertRaises(SchemaValidationException) as error:
            check_class(
                {
                    "class": [],
                }
            )
        check_error_message(self, error, key_type_error_messsage("class", [], str))

        with self.assertRaises(SchemaValidationException) as error:
            check_class({"class": "Tree", "invalid_key": []})
        check_error_message(self, error, unknown_key_error_message("invalid_key"))

        with self.assertRaises(SchemaValidationException) as error:
            check_class({"class": "Tree", "description": []})
        check_error_message(self, error, key_type_error_messsage("description", [], str))

        with self.assertRaises(SchemaValidationException) as error:
            check_class({"class": "Tree", "properties": "References please"})
        check_error_message(self, error, key_type_error_messsage("properties", "", list))

        with self.assertRaises(SchemaValidationException) as error:
            check_class({"class": "Tree", "vectorIndexType": True})
        check_error_message(self, error, key_type_error_messsage("vectorIndexType", True, str))

        with self.assertRaises(SchemaValidationException) as error:
            check_class({"class": "Tree", "vectorIndexConfig": []})
        check_error_message(self, error, key_type_error_messsage("vectorIndexConfig", [], dict))

        with self.assertRaises(SchemaValidationException) as error:
            check_class({"class": "Tree", "replicationConfig": []})
        check_error_message(self, error, key_type_error_messsage("replicationConfig", [], dict))

        with self.assertRaises(SchemaValidationException) as error:
            check_class({"class": "Tree", "moduleConfig": []})
        check_error_message(self, error, key_type_error_messsage("moduleConfig", [], dict))

        with self.assertRaises(SchemaValidationException) as error:
            check_class({"class": "Tree", "vectorizer": 100.1})
        check_error_message(self, error, key_type_error_messsage("vectorizer", 100.1, str))

        # check if `check_property` is called inside `check_class`
        with self.assertRaises(SchemaValidationException) as error:
            check_class(
                {
                    "class": "Tree",
                    "properties": [
                        {
                            "dataType": ["string"],
                            "description": "Test Property",
                        }
                    ],
                }
            )
        check_error_message(self, error, 'Property does not contain "name"')

    def test_check_property(self):
        """
        Test `check_property` function.
        """

        # valid calls
        valid_minimal = {"dataType": ["string"], "name": "string"}

        check_property(valid_minimal)
        valid_max = {
            "dataType": ["string"],
            "name": "Rocket",
            "moduleConfig": {},
            "description": "some description",
            "indexInverted": True,
        }
        check_property(valid_max)

        # invalid calls
        data_type_error_message = 'Property does not contain "dataType"'
        name_error_message = 'Property does not contain "name"'
        key_error_message = lambda key: f'Property "{key}" is not known.'
        key_type_error_messsage = lambda key, value, exp_type: (
            f'"{key}" is type {type(value)} ' f"but should be {exp_type}."
        )

        properties = {"dataType": ["string"]}
        with self.assertRaises(SchemaValidationException) as error:
            check_property(properties)
        check_error_message(self, error, name_error_message)

        properties = {"name": "string"}
        with self.assertRaises(SchemaValidationException) as error:
            check_property(properties)
        check_error_message(self, error, data_type_error_message)

        properties = {"dataType": ["string"], "name": "string", "invalid_property": "value"}
        with self.assertRaises(SchemaValidationException) as error:
            check_property(properties)
        check_error_message(self, error, key_error_message("invalid_property"))

        properties = {
            "dataType": ["string"],
            "name": "Rocket",
            "moduleConfig": [],
        }
        with self.assertRaises(SchemaValidationException) as error:
            check_property(properties)
        check_error_message(
            self, error, key_type_error_messsage("moduleConfig", properties["moduleConfig"], dict)
        )

        properties = {
            "dataType": ["string"],
            "name": "Rocket",
            "description": ["some description"],
        }
        with self.assertRaises(SchemaValidationException) as error:
            check_property(properties)
        check_error_message(
            self, error, key_type_error_messsage("description", properties["description"], str)
        )

        properties = {"dataType": ["string"], "name": "Rocket", "indexInverted": "True"}
        with self.assertRaises(SchemaValidationException) as error:
            check_property(properties)
        check_error_message(
            self, error, key_type_error_messsage("indexInverted", properties["indexInverted"], bool)
        )

        properties = {"dataType": ["string", 10], "name": "Rocket", "indexInverted": True}
        with self.assertRaises(SchemaValidationException) as error:
            check_property(properties)
        check_error_message(
            self, error, key_type_error_messsage("dataType object", properties["dataType"][1], str)
        )



import unittest
from numbers import Real
from unittest.mock import Mock, patch

import pytest
from requests import ReadTimeout
from requests.exceptions import ConnectionError as RequestsConnectionError

from test.util import mock_connection_func, check_error_message, check_startswith_error_message
from weaviate.batch import Batch
from weaviate.batch.crud_batch import WeaviateErrorRetryConf
from weaviate.batch.requests import ObjectsBatchRequest, ReferenceBatchRequest
from weaviate.data.replication import ConsistencyLevel
from weaviate.exceptions import UnexpectedStatusCodeException


@pytest.mark.parametrize(
    "num,excl,incl",
    [
        (3, ["some entry"], ["other entry"]),
        (-1, None, None),
        (1, ["some entry", 1], None),
        (1, None, ["some entry", 1]),
        (1, None, []),
    ],
)
def test_retry_config_both_lists(num, excl, incl):
    """Test that either include or exclude list can be given."""
    with pytest.raises(ValueError):
        WeaviateErrorRetryConf(number_retries=num, errors_to_exclude=excl, errors_to_include=incl)


class TestBatch(unittest.TestCase):
    def check_instance(
        self,
        batch: Batch,
        recom_num_obj: int = None,
        recom_num_ref: int = None,
        batch_size: int = None,
        creation_time: Real = 2,  # this is based on the default timeout_config of the Client class
        timeout_retries: int = 3,
        connection_error_retries: int = 3,
        batching_type: str = None,
        num_workers: int = 1,
        consistency_level: ConsistencyLevel = None,
    ) -> None:
        """
        Check all configurable attributes of the Batch instance.
        """

        self.assertEqual(batch._recommended_num_objects, recom_num_obj)
        self.assertEqual(batch._recommended_num_references, recom_num_ref)
        self.assertEqual(batch._batch_size, batch_size)
        self.assertEqual(batch._creation_time, creation_time)
        self.assertEqual(batch._timeout_retries, timeout_retries)
        self.assertEqual(batch._connection_error_retries, connection_error_retries)
        self.assertEqual(batch._batching_type, batching_type)
        self.assertEqual(batch._num_workers, num_workers)
        self.assertEqual(batch._consistency_level, consistency_level)

    # TEST SETTERS/GETTERS

    def test_timeout_retries(self):
        """
        Test Setter and Getter for 'timeout_retries'.
        """

        batch = Batch(mock_connection_func())
        self.check_instance(batch=batch)

        self.assertEqual(batch.timeout_retries, 3)
        self.check_instance(batch, timeout_retries=3)

        batch.timeout_retries = 10
        self.assertEqual(batch.timeout_retries, 10)
        self.check_instance(batch, timeout_retries=10)

        batch.timeout_retries = 0
        self.assertEqual(batch.timeout_retries, 0)
        self.check_instance(batch, timeout_retries=0)

        batch.timeout_retries = 1
        self.assertEqual(batch.timeout_retries, 1)
        self.check_instance(batch, timeout_retries=1)

        # exceptions
        ## error messages
        value_error = "'timeout_retries' must be positive, i.e. greater or equal that zero (>=0)."
        type_error = f"'timeout_retries' must be of type {int}."

        #######################################################################
        # test wrong value
        with self.assertRaises(ValueError) as error:
            batch.timeout_retries = -1
        self.assertEqual(batch.timeout_retries, 1)
        self.check_instance(batch, timeout_retries=1)
        check_error_message(self, error, value_error)

        #######################################################################
        # test wrong type
        with self.assertRaises(TypeError) as error:
            batch.timeout_retries = True
        self.assertEqual(batch.timeout_retries, 1)
        self.check_instance(batch, timeout_retries=1)
        check_error_message(self, error, type_error)

        with self.assertRaises(TypeError) as error:
            batch.timeout_retries = "2"
        self.assertEqual(batch.timeout_retries, 1)
        self.check_instance(batch, timeout_retries=1)
        check_error_message(self, error, type_error)

    def test_connection_error_retries(self):
        """
        Test Setter and Getter for 'connection_error_retries'.
        """

        batch = Batch(mock_connection_func())
        self.check_instance(batch=batch)

        self.assertEqual(batch.connection_error_retries, 3)
        self.check_instance(batch, connection_error_retries=3)

        batch.connection_error_retries = 10
        self.assertEqual(batch.connection_error_retries, 10)
        self.check_instance(batch, connection_error_retries=10)

        batch.connection_error_retries = 0
        self.assertEqual(batch.connection_error_retries, 0)
        self.check_instance(batch, connection_error_retries=0)

        batch.connection_error_retries = 1
        self.assertEqual(batch.connection_error_retries, 1)
        self.check_instance(batch, connection_error_retries=1)

        # exceptions
        ## error messages
        value_error = (
            "'connection_error_retries' must be positive, i.e. greater or equal that zero (>=0)."
        )
        type_error = f"'connection_error_retries' must be of type {int}."

        #######################################################################
        # test wrong value
        with self.assertRaises(ValueError) as error:
            batch.connection_error_retries = -1
        self.assertEqual(batch.connection_error_retries, 1)
        self.check_instance(batch, connection_error_retries=1)
        check_error_message(self, error, value_error)

        #######################################################################
        # test wrong type
        with self.assertRaises(TypeError) as error:
            batch.connection_error_retries = True
        self.assertEqual(batch.connection_error_retries, 1)
        self.check_instance(batch, connection_error_retries=1)
        check_error_message(self, error, type_error)

        with self.assertRaises(TypeError) as error:
            batch.connection_error_retries = "2"
        self.assertEqual(batch.connection_error_retries, 1)
        self.check_instance(batch, connection_error_retries=1)
        check_error_message(self, error, type_error)

    def test_recommended_num_objects(self):
        """
        Test Setter and Getter for 'recommended_num_objects'.
        """

        batch = Batch(mock_connection_func())
        self.check_instance(batch)

        self.assertIsNone(batch.recommended_num_objects)
        self.check_instance(batch)

        batch._recommended_num_objects = 10
        self.assertEqual(batch.recommended_num_objects, 10)
        self.check_instance(batch, recom_num_obj=10)

        batch._recommended_num_objects = 20
        self.assertEqual(batch.recommended_num_objects, 20)
        self.check_instance(batch, recom_num_obj=20)

    def test_recommended_num_references(self):
        """
        Test Setter and Getter for 'recommended_num_references'.
        """

        batch = Batch(mock_connection_func())
        self.check_instance(batch)

        self.assertIsNone(batch.recommended_num_references)
        self.check_instance(batch)

        batch._recommended_num_references = 10
        self.assertEqual(batch.recommended_num_references, 10)
        self.check_instance(batch, recom_num_ref=10)

        batch._recommended_num_references = 20
        self.assertEqual(batch.recommended_num_references, 20)
        self.check_instance(batch, recom_num_ref=20)

    @patch("weaviate.batch.crud_batch.Batch._auto_create")
    def test_batch_size(self, mock_auto_create):
        """
        Test Setter and Getter for 'batch_size'.
        """

        batch = Batch(mock_connection_func())
        self.check_instance(batch)

        self.assertIsNone(batch.batch_size)
        self.check_instance(batch)

        #######################################################################
        # test batch_size: None -> int
        batch.batch_size = 10
        self.assertEqual(batch.batch_size, 10)
        self.check_instance(
            batch, batch_size=10, batching_type="fixed", recom_num_obj=10, recom_num_ref=10
        )
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

        #######################################################################
        # test batch_size: int -> int and dynamic enabled, one recommended set to None
        ## set some attributes manually (ONLY WHEN TESTING!!!)
        batch._batching_type = "dynamic"
        batch._recommended_num_objects = None

        batch.batch_size = 200
        self.assertEqual(batch.batch_size, 200)
        self.check_instance(
            batch, batch_size=200, batching_type="dynamic", recom_num_obj=200, recom_num_ref=10
        )
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

        #######################################################################
        # test batch_size: int -> None
        batch.batch_size = None
        self.assertIsNone(batch.batch_size)
        self.check_instance(
            batch, batch_size=None, batching_type=None, recom_num_obj=200, recom_num_ref=10
        )
        mock_auto_create.assert_not_called()
        mock_auto_create.reset_mock()

        #######################################################################
        # test exceptions
        ## messages
        type_error = f"'batch_size' must be of type {int}."
        value_error = "'batch_size' must be positive, i.e. greater that zero (>0)."

        with self.assertRaises(TypeError) as error:
            batch.batch_size = False
        check_error_message(self, error, type_error)
        self.check_instance(
            batch, batch_size=None, batching_type=None, recom_num_obj=200, recom_num_ref=10
        )
        mock_auto_create.assert_not_called()

        with self.assertRaises(TypeError) as error:
            batch.batch_size = 100.5
        check_error_message(self, error, type_error)
        self.check_instance(
            batch, batch_size=None, batching_type=None, recom_num_obj=200, recom_num_ref=10
        )
        mock_auto_create.assert_not_called()

        with self.assertRaises(ValueError) as error:
            batch.batch_size = 0
        check_error_message(self, error, value_error)
        self.check_instance(
            batch, batch_size=None, batching_type=None, recom_num_obj=200, recom_num_ref=10
        )
        mock_auto_create.assert_not_called()

        with self.assertRaises(ValueError) as error:
            batch.batch_size = -100
        check_error_message(self, error, value_error)
        self.check_instance(
            batch, batch_size=None, batching_type=None, recom_num_obj=200, recom_num_ref=10
        )
        mock_auto_create.assert_not_called()

    @patch("weaviate.batch.crud_batch.Batch._auto_create")
    def test_dynamic(self, mock_auto_create):
        """
        Test Setter and Getter for 'dynamic'.
        """

        batch = Batch(mock_connection_func())
        self.check_instance(batch)

        self.assertFalse(batch.dynamic)
        self.check_instance(batch)

        #######################################################################
        # test when batching_type is None (setter does nothing)
        batch.dynamic = True
        self.assertFalse(batch.dynamic)
        self.check_instance(batch, batching_type=None)

        #######################################################################
        # test when batching_type not None
        batch._batching_type = "fixed"
        batch.dynamic = True
        self.assertTrue(batch.dynamic)
        self.check_instance(batch, batching_type="dynamic")
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

        # set again to True
        batch.dynamic = True
        self.assertTrue(batch.dynamic)
        self.check_instance(batch, batching_type="dynamic")
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

        batch.dynamic = False
        self.assertFalse(batch.dynamic)
        self.check_instance(batch, batching_type="fixed")
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

        #######################################################################
        # test exceptions
        ## messages
        type_error = "'dynamic' must be of type bool."

        with self.assertRaises(TypeError) as error:
            batch.dynamic = 0
        check_error_message(self, error, type_error)
        self.check_instance(batch, batching_type="fixed")
        mock_auto_create.assert_not_called()

    @patch("weaviate.batch.crud_batch.Batch._auto_create")
    def test_creation_time(self, mock_auto_create):
        """
        Test Setter and Getter for 'creation_time'.
        """

        batch = Batch(mock_connection_func())
        self.check_instance(batch)

        self.assertEqual(batch.creation_time, 2)
        self.check_instance(batch)

        #######################################################################
        # test when batching_type and recommended are None
        batch.creation_time = 5.0
        self.assertEqual(batch.creation_time, 5.0)
        self.check_instance(batch, creation_time=5.0)
        mock_auto_create.assert_not_called()

        #######################################################################
        # test when recommended are not None, but batching_type is None
        batch._recommended_num_objects = 100
        batch._recommended_num_references = 200
        batch.creation_time = 20.0
        self.assertEqual(batch.creation_time, 20.0)
        self.check_instance(
            batch,
            creation_time=20.0,
            batching_type=None,
            recom_num_obj=400,
            recom_num_ref=800,
        )
        mock_auto_create.assert_not_called()

        #######################################################################
        # test when recommended are None
        batch.batch_size = 10  # sets batching_type to 'fixed'
        batch.creation_time = 5.0
        self.assertEqual(batch.creation_time, 5.0)
        self.check_instance(
            batch,
            batch_size=10,
            creation_time=5.0,
            batching_type="fixed",
            recom_num_obj=100,
            recom_num_ref=200,
        )
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

        #######################################################################
        # test exceptions
        ## messages
        type_error = f"'creation_time' must be of type {Real}."
        value_error = "'creation_time' must be positive, i.e. greater that zero (>0)."

        with self.assertRaises(TypeError) as error:
            batch.creation_time = True
        check_error_message(self, error, type_error)
        self.check_instance(
            batch,
            batch_size=10,
            creation_time=5.0,
            batching_type="fixed",
            recom_num_obj=100,
            recom_num_ref=200,
        )
        mock_auto_create.assert_not_called()

        with self.assertRaises(TypeError) as error:
            batch.creation_time = "10.6"
        check_error_message(self, error, type_error)
        self.check_instance(
            batch,
            batch_size=10,
            creation_time=5.0,
            batching_type="fixed",
            recom_num_obj=100,
            recom_num_ref=200,
        )
        mock_auto_create.assert_not_called()

        with self.assertRaises(ValueError) as error:
            batch.creation_time = 0.0
        check_error_message(self, error, value_error)
        self.check_instance(
            batch,
            batch_size=10,
            creation_time=5.0,
            batching_type="fixed",
            recom_num_obj=100,
            recom_num_ref=200,
        )
        mock_auto_create.assert_not_called()

        with self.assertRaises(ValueError) as error:
            batch.creation_time = -10.0
        check_error_message(self, error, value_error)
        self.check_instance(
            batch,
            batch_size=10,
            creation_time=5.0,
            batching_type="fixed",
            recom_num_obj=100,
            recom_num_ref=200,
        )
        mock_auto_create.assert_not_called()

    def test_shape_and_num_data(self):
        """
        Test the `shape`, `num_objects` and `num_references` property/methods.
        """

        batch = Batch(mock_connection_func(server_version="1.14.0"))

        self.assertEqual(batch.num_objects(), 0)
        self.assertEqual(batch.num_references(), 0)
        self.assertEqual(batch.shape, (0, 0))

        #######################################################################
        # add one object
        batch.add_data_object({}, "Test")

        self.assertEqual(batch.num_objects(), 1)
        self.assertEqual(batch.num_references(), 0)
        self.assertEqual(batch.shape, (1, 0))

        #######################################################################
        # add one object
        batch.add_data_object({}, "Test")

        self.assertEqual(batch.num_objects(), 2)
        self.assertEqual(batch.num_references(), 0)
        self.assertEqual(batch.shape, (2, 0))

        #######################################################################
        # add one reference
        batch.add_reference(
            "f0153f24-3923-4046-919b-6a3e8fd37394",
            "Test",
            "test",
            "f0153f24-3923-4046-919b-6a3e8fd37395",
        )

        self.assertEqual(batch.num_objects(), 2)
        self.assertEqual(batch.num_references(), 1)
        self.assertEqual(batch.shape, (2, 1))

        #######################################################################
        # add one reference
        batch.add_reference(
            "f0153f24-3923-4046-919b-6a3e8fd37396",
            "Test",
            "test",
            "f0153f24-3923-4046-919b-6a3e8fd37397",
        )

        self.assertEqual(batch.num_objects(), 2)
        self.assertEqual(batch.num_references(), 2)
        self.assertEqual(batch.shape, (2, 2))

    @patch("weaviate.batch.crud_batch.Batch.shutdown")
    @patch("weaviate.batch.crud_batch.Batch.start")
    @patch("weaviate.batch.crud_batch.Batch.flush")
    def test_enter_exit(self, mock_flush, mock_start, mock_shutdown):
        """
        Test `__enter__` and `__exit__` methods.
        """

        mock_start.return_value = "test"
        batch = Batch(mock_connection_func())

        with batch as b:
            self.assertEqual("test", b)
            mock_flush.assert_not_called()
            mock_shutdown.assert_not_called()
            mock_start.assert_called()
            mock_start.reset_mock()
        mock_flush.assert_called()
        mock_shutdown.assert_called()
        mock_start.assert_not_called()

    def test_shutdown(
        self,
    ):
        """
        Test `shutdown` method.
        """

        # initial value for executor (None)
        batch = Batch(mock_connection_func())
        with patch("weaviate.batch.crud_batch.BatchExecutor") as mock_executor:
            mock_executor.shutdown.assert_not_called()
            batch.shutdown()
            mock_executor.shutdown.assert_not_called()

        # shutdown executor
        batch = Batch(mock_connection_func())
        mock_executor = Mock()
        mock_executor.is_shutdown.return_value = True
        batch._executor = mock_executor
        mock_executor.shutdown.assert_not_called()
        mock_executor.is_shutdown.assert_not_called()
        batch.shutdown()
        mock_executor.is_shutdown.assert_called()
        mock_executor.shutdown.assert_not_called()

        # running executor
        batch = Batch(mock_connection_func())
        mock_executor = Mock()
        mock_executor.is_shutdown.return_value = False
        batch._executor = mock_executor
        mock_executor.shutdown.assert_not_called()
        mock_executor.is_shutdown.assert_not_called()
        batch.shutdown()
        mock_executor.is_shutdown.assert_called()
        mock_executor.shutdown.assert_called()

    @patch("weaviate.batch.crud_batch.BatchExecutor")
    def test_start(self, mock_executor_class):
        """
        Test `start` method.
        """

        executor = Mock()
        executor.is_shutdown.return_value = False
        mock_executor_class.return_value = executor

        batch = Batch(mock_connection_func())
        mock_executor_class.assert_not_called()

        mock_executor_class.reset_mock()
        mock_executor_class.return_value = executor
        executor.is_shutdown.assert_not_called()
        result = batch.start()
        self.assertEqual(result, batch)
        executor.is_shutdown.assert_not_called()
        mock_executor_class.assert_called()

        executor.reset_mock()

        mock_executor_class.reset_mock()
        mock_executor_class.return_value = executor
        executor.is_shutdown.assert_not_called()
        result = batch.start()
        self.assertEqual(result, batch)
        executor.is_shutdown.assert_called()
        mock_executor_class.assert_not_called()

        executor.reset_mock()

        executor.is_shutdown.return_value = True
        batch._num_workers = 2
        result = batch.start()
        self.assertEqual(result, batch)
        executor.is_shutdown.assert_called()
        mock_executor_class.assert_called_with(max_workers=2)

    @patch("weaviate.batch.crud_batch.Batch._send_batch_requests")
    def test_flush(self, mock_send_batch_requests):
        """
        Test `flush` method.
        """

        batch = Batch(mock_connection_func())

        batch.flush()
        mock_send_batch_requests.assert_called_with(force_wait=True)
        mock_send_batch_requests.reset_mock()

    @patch("weaviate.batch.crud_batch.Batch._send_batch_requests")
    def test_auto_create(self, mock_send_batch_requests):
        """
        Test `_auto_create` method through `add_*` methods.
        """

        batch = Batch(mock_connection_func(server_version="1.14.0"))
        batch.batch_size = 2  # this enables auto_create with batching `fixed`

        #######################################################################
        # only objects and batching_type 'fixed'
        batch.add_data_object({}, "Test")
        mock_send_batch_requests.assert_not_called()
        batch.add_data_object({}, "Test")
        mock_send_batch_requests.assert_called()
        mock_send_batch_requests.reset_mock()

        #######################################################################
        # add references too, batching_type 'fixed'
        batch.batch_size = 4
        batch.add_reference(
            "f0153f24-3923-4046-919b-6a3e8fd37393",
            "Test",
            "test",
            "f0153f24-3923-4046-919b-6a3e8fd37394",
        )
        mock_send_batch_requests.assert_not_called()
        batch.add_reference(
            "f0153f24-3923-4046-919b-6a3e8fd37396",
            "Test",
            "test",
            "f0153f24-3923-4046-919b-6a3e8fd37397",
        )
        mock_send_batch_requests.assert_called()
        mock_send_batch_requests.reset_mock()

        #######################################################################
        # batching_type 'dynamic'
        batch = Batch(mock_connection_func(server_version="1.14.0"))
        batch.batch_size = 2  # this enables auto_create with batching `fixed`
        batch.dynamic = True  # NOTE: recommended are set to 2 when we set the batch_size to 2
        batch.add_data_object({}, "Test")
        mock_send_batch_requests.assert_not_called()
        batch.add_reference(
            "f0153f24-3923-4046-919b-6a3e8fd37393",
            "Test",
            "test",
            "f0153f24-3923-4046-919b-6a3e8fd37394",
        )
        mock_send_batch_requests.assert_not_called()
        batch.add_reference(
            "f0153f24-3923-4046-919b-6a3e8fd37396",
            "Test",
            "test",
            "f0153f24-3923-4046-919b-6a3e8fd37397",
        )
        mock_send_batch_requests.assert_called()
        batch.add_data_object({}, "Test")
        mock_send_batch_requests.assert_called()
        mock_send_batch_requests.reset_mock()

        #######################################################################
        # exceptions
        ## error messages
        value_error = f'Unsupported batching type "{None}"'
        batch = Batch(mock_connection_func())
        with self.assertRaises(ValueError) as error:
            batch._auto_create()  # This should not be called like this, only for test purposes
        check_error_message(self, error, value_error)

    @patch("weaviate.batch.crud_batch.Batch._auto_create")
    def test_add_data_object(self, mock_auto_create):
        """
        Test `add_data_object` method.
        """

        batch = Batch(mock_connection_func())
        batch._objects_batch = Mock()  # to test if the add method is called

        batch.add_data_object({}, "Test")
        self.assertEqual(batch._objects_batch.add.call_count, 1)
        batch._objects_batch.add.assert_called_with(
            class_name="Test",
            data_object={},
            uuid=None,
            vector=None,
        )
        mock_auto_create.assert_not_called()

        batch.add_data_object({}, "Test")
        self.assertEqual(batch._objects_batch.add.call_count, 2)
        batch._objects_batch.add.assert_called_with(
            class_name="Test",
            data_object={},
            uuid=None,
            vector=None,
        )
        mock_auto_create.assert_not_called()

        batch._batching_type = (
            "fixed"  # This should not be called like this, only for test purposes
        )
        batch.add_data_object({}, "Test")
        self.assertEqual(batch._objects_batch.add.call_count, 3)
        batch._objects_batch.add.assert_called_with(
            class_name="Test",
            data_object={},
            uuid=None,
            vector=None,
        )
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

        batch._batching_type = (
            "dynamic"  # This should not be called like this, only for test purposes
        )
        batch.add_data_object({}, "Test")
        self.assertEqual(batch._objects_batch.add.call_count, 4)
        batch._objects_batch.add.assert_called_with(
            class_name="Test",
            data_object={},
            uuid=None,
            vector=None,
        )
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

        batch.add_data_object({}, "test")
        self.assertEqual(batch._objects_batch.add.call_count, 5)
        batch._objects_batch.add.assert_called_with(
            class_name="Test",
            data_object={},
            uuid=None,
            vector=None,
        )
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

    @patch("weaviate.batch.crud_batch.Batch._auto_create")
    def test_add_reference(self, mock_auto_create):
        """
        Test `add_reference` method.
        """

        batch = Batch(mock_connection_func(server_version="1.14.0"))
        batch._reference_batch = Mock()  # to test if the add method is called

        batch.add_reference(
            "f0153f24-3923-4046-919b-6a3e8fd37391",
            "Test",
            "test",
            "f0153f24-3923-4046-919b-6a3e8fd37392",
        )
        self.assertEqual(batch._reference_batch.add.call_count, 1)
        batch._reference_batch.add.assert_called_with(
            from_object_class_name="Test",
            from_object_uuid="f0153f24-3923-4046-919b-6a3e8fd37391",
            from_property_name="test",
            to_object_uuid="f0153f24-3923-4046-919b-6a3e8fd37392",
            to_object_class_name=None,
        )
        mock_auto_create.assert_not_called()

        batch.add_reference(
            "f0153f24-3923-4046-919b-6a3e8fd37393",
            "Test",
            "test",
            "f0153f24-3923-4046-919b-6a3e8fd37394",
        )
        self.assertEqual(batch._reference_batch.add.call_count, 2)
        batch._reference_batch.add.assert_called_with(
            from_object_class_name="Test",
            from_object_uuid="f0153f24-3923-4046-919b-6a3e8fd37393",
            from_property_name="test",
            to_object_uuid="f0153f24-3923-4046-919b-6a3e8fd37394",
            to_object_class_name=None,
        )
        mock_auto_create.assert_not_called()

        batch._batching_type = (
            "fixed"  # This should never be called like this, we do it for test purposes
        )
        batch.add_reference(
            "f0153f24-3923-4046-919b-6a3e8fd37396",
            "Test",
            "test",
            "f0153f24-3923-4046-919b-6a3e8fd37397",
        )
        self.assertEqual(batch._reference_batch.add.call_count, 3)
        batch._reference_batch.add.assert_called_with(
            from_object_class_name="Test",
            from_object_uuid="f0153f24-3923-4046-919b-6a3e8fd37396",
            from_property_name="test",
            to_object_uuid="f0153f24-3923-4046-919b-6a3e8fd37397",
            to_object_class_name=None,
        )
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

        batch._batching_type = (
            "dynamic"  # This should never be called like this, we do it for test purposes
        )
        batch.add_reference(
            "f0153f24-3923-4046-919b-6a3e8fd37390",
            "Test",
            "test",
            "f0153f24-3923-4046-919b-6a3e8fd37399",
        )
        self.assertEqual(batch._reference_batch.add.call_count, 4)
        batch._reference_batch.add.assert_called_with(
            from_object_class_name="Test",
            from_object_uuid="f0153f24-3923-4046-919b-6a3e8fd37390",
            from_property_name="test",
            to_object_uuid="f0153f24-3923-4046-919b-6a3e8fd37399",
            to_object_class_name=None,
        )
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

        batch.add_reference(
            "f0153f24-3923-4046-919b-6a3e8fd37395",
            "test",
            "test",
            "f0153f24-3923-4046-919b-6a3e8fd37319",
        )
        self.assertEqual(batch._reference_batch.add.call_count, 5)
        batch._reference_batch.add.assert_called_with(
            from_object_class_name="Test",
            from_object_uuid="f0153f24-3923-4046-919b-6a3e8fd37395",
            from_property_name="test",
            to_object_uuid="f0153f24-3923-4046-919b-6a3e8fd37319",
            to_object_class_name=None,
        )
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

    @patch("weaviate.batch.crud_batch.Batch._create_data")
    def test_create_objects(self, mock_create_data):
        """
        Test `create_objects` method.
        """

        batch = Batch(mock_connection_func())
        ## mock the requests.Response object
        mock_response = Mock()
        mock_response.json.return_value = "Test"
        mock_response.elapsed.total_seconds.return_value = 1.0
        mock_create_data.return_value = mock_response

        #######################################################################
        # create zero length batch
        self.assertEqual(batch.create_objects(), [])
        mock_create_data.assert_not_called()
        self.check_instance(batch)

        #######################################################################
        # add objects
        batch.add_data_object({}, "Test")
        batch.add_data_object({}, "Test")
        self.assertEqual(batch.create_objects(), "Test")
        mock_create_data.assert_called()
        self.check_instance(batch, recom_num_obj=0)
        self.assertEqual(batch.num_objects(), 0)

    @patch("weaviate.batch.crud_batch.Batch._create_data")
    def test_create_references(self, mock_create_data):
        """
        Test `create_references` method.
        """

        batch = Batch(mock_connection_func(server_version="1.14.0"))
        ## mock the requests.Response object
        mock_response = Mock()
        mock_response.json.return_value = "Test"
        mock_response.elapsed.total_seconds.return_value = 1.0
        mock_create_data.return_value = mock_response

        #######################################################################
        # create zero length batch
        self.assertEqual(batch.create_references(), [])
        mock_create_data.assert_not_called()
        self.check_instance(batch)

        #######################################################################
        # add objects
        batch.add_reference(
            "f0153f24-3923-4046-919b-6a3e8fd37391",
            "Test",
            "test",
            "f0153f24-3923-4046-919b-6a3e8fd37398",
        )
        batch.add_reference(
            "f0153f24-3923-4046-919b-6a3e8fd37390",
            "Test",
            "test",
            "f0153f24-3923-4046-919b-6a3e8fd37399",
        )
        self.assertEqual(batch.create_references(), "Test")
        mock_create_data.assert_called()
        self.check_instance(batch, recom_num_ref=0)
        self.assertEqual(batch.num_references(), 0)

    def test_create_data(self):
        """
        Test the `_create_data` method.
        """

        #######################################################################
        # test status_code == 200, timeout_retries = 0
        mock_connection = mock_connection_func("post", status_code=200)
        batch = Batch(mock_connection)
        batch._create_data("references", ReferenceBatchRequest())
        mock_connection.post.assert_called_with(
            path="/batch/references",
            weaviate_object=[],
            params=None,
        )
        self.assertEqual(mock_connection.post.call_count, 1)

        #######################################################################
        # timeout_retries = 2, and no exception raised
        mock_connection = mock_connection_func("post", status_code=200)
        batch = Batch(mock_connection)
        batch.timeout_retries = 2
        batch.consistency_level = ConsistencyLevel.QUORUM
        batch._create_data("references", ReferenceBatchRequest())
        mock_connection.post.assert_called_with(
            path="/batch/references",
            weaviate_object=[],
            params={"consistency_level": "QUORUM"},
        )
        self.assertEqual(mock_connection.post.call_count, 1)

        #######################################################################
        # test errors
        #######################################################################
        ## error messages
        requests_error_message = "Batch was not added to weaviate."
        unexpected_error_message = lambda data: f"Create {data} in batch"

        #######################################################################
        ## test RequestsConnectionError
        mock_connection = mock_connection_func("post", side_effect=RequestsConnectionError("Test!"))
        batch = Batch(mock_connection)
        batch.consistency_level = ConsistencyLevel.ONE
        with self.assertRaises(RequestsConnectionError) as error:
            batch._create_data("objects", ObjectsBatchRequest())
        check_error_message(self, error, requests_error_message)
        mock_connection.post.assert_called_with(
            path="/batch/objects",
            weaviate_object={"fields": ["ALL"], "objects": []},
            params={"consistency_level": "ONE"},
        )

        ## test ConnectionError, connection_error_retries = 0
        mock_connection = mock_connection_func("post", side_effect=RequestsConnectionError("Test!"))
        mock_connection.timeout_config = (2, 100)
        batch = Batch(mock_connection)
        batch.connection_error_retries = 0
        batch.consistency_level = ConsistencyLevel.ALL
        with self.assertRaises(RequestsConnectionError) as error:
            batch._create_data("references", ReferenceBatchRequest())
        check_startswith_error_message(self, error, requests_error_message)
        mock_connection.post.assert_called_with(
            path="/batch/references", weaviate_object=[], params={"consistency_level": "ALL"}
        )
        self.assertEqual(mock_connection.post.call_count, 1)

        ## test ConnectionError, connection_error_retries = 3 (default)
        mock_connection = mock_connection_func("post", side_effect=RequestsConnectionError("Test!"))
        mock_connection.timeout_config = (2, 100)
        batch = Batch(mock_connection)
        with self.assertRaises(RequestsConnectionError) as error:
            batch._create_data("objects", ObjectsBatchRequest())
        check_startswith_error_message(self, error, requests_error_message)
        mock_connection.post.assert_called_with(
            path="/batch/objects",
            weaviate_object={"fields": ["ALL"], "objects": []},
            params=None,
        )
        self.assertEqual(mock_connection.post.call_count, 3 + 1)

        ## test alternating errors
        i_for_alt_errors = 0

        def alternating_errors():
            nonlocal i_for_alt_errors
            if i_for_alt_errors % 2 == 0:
                err = RequestsConnectionError
            else:
                err = ReadTimeout
            i_for_alt_errors += 1
            raise err

        ## test status_code != 200
        mock_connection = mock_connection_func("post", status_code=204)
        batch = Batch(mock_connection)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            batch._create_data("references", ReferenceBatchRequest())
        check_startswith_error_message(self, error, unexpected_error_message("references"))
        mock_connection.post.assert_called_with(
            path="/batch/references",
            weaviate_object=[],
            params=None,
        )

        batch = Batch(mock_connection)
        with self.assertRaises(ValueError) as error:
            batch.consistency_level = 1
        check_startswith_error_message(self, error, "1 is not a valid ConsistencyLevel")

    @patch("weaviate.batch.crud_batch.Batch._auto_create")
    def test_configure_call(self, mock_auto_create):
        """
        Test the `configure` method, which is the same as `__call__`.
        """

        batch = Batch(mock_connection_func())
        self.check_instance(batch)
        #
        with self.assertRaises(ValueError) as error:
            batch.configure(consistency_level=1)
        check_startswith_error_message(self, error, "1 is not a valid ConsistencyLevel")
        #######################################################################
        # batching_type: None -> 'fixed'
        return_batch = batch.configure(
            batch_size=100,
            creation_time=20.76,
            timeout_retries=2,
            consistency_level=ConsistencyLevel.ALL,
        )
        self.assertEqual(batch, return_batch)
        self.check_instance(
            batch,
            batch_size=100,
            creation_time=20.76,
            timeout_retries=2,
            batching_type="fixed",
            consistency_level="ALL",
        )
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

        #######################################################################
        # batching_type: 'fixed' -> 'dynamic'
        return_batch = batch.configure(
            batch_size=200,
            creation_time=2.5,
            timeout_retries=0,
            dynamic=True,
        )
        self.assertEqual(batch, return_batch)
        self.check_instance(
            batch,
            batch_size=200,
            creation_time=2.5,
            timeout_retries=0,
            batching_type="dynamic",
            recom_num_obj=200,
            recom_num_ref=200,
        )
        mock_auto_create.assert_called()
        mock_auto_create.reset_mock()

        #######################################################################
        # batching_type: 'dynamic' -> None
        return_batch = batch.configure(
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            dynamic=True,
            consistency_level="QUORUM",
        )
        self.assertEqual(batch, return_batch)
        self.check_instance(
            batch,
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            batching_type=None,
            recom_num_ref=200,  # does not change if not None
            recom_num_obj=200,  # does not change if not None
            consistency_level=ConsistencyLevel.QUORUM,
        )
        mock_auto_create.assert_not_called()

        #######################################################################
        # test errors
        #######################################################################

        #######################################################################
        # creation_time

        type_error = f"'creation_time' must be of type {Real}."
        value_error = "'creation_time' must be positive, i.e. greater that zero (>0)."

        with self.assertRaises(TypeError) as error:
            batch.configure(
                batch_size=None,
                creation_time=True,
                timeout_retries=10,
                dynamic=True,
            )

        check_error_message(self, error, type_error)
        self.check_instance(
            batch,
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            batching_type=None,
            recom_num_ref=200,  # does not change if not None
            recom_num_obj=200,  # does not change if not None
        )
        mock_auto_create.assert_not_called()

        with self.assertRaises(TypeError) as error:
            batch.configure(
                batch_size=None,
                creation_time="12.5",
                timeout_retries=10,
                dynamic=True,
            )
        check_error_message(self, error, type_error)
        self.check_instance(
            batch,
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            batching_type=None,
            recom_num_ref=200,  # does not change if not None
            recom_num_obj=200,  # does not change if not None
        )
        mock_auto_create.assert_not_called()

        with self.assertRaises(ValueError) as error:
            batch.configure(
                batch_size=None,
                creation_time=0.0,
                timeout_retries=10,
                dynamic=True,
            )
        check_error_message(self, error, value_error)
        self.check_instance(
            batch,
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            batching_type=None,
            recom_num_ref=200,  # does not change if not None
            recom_num_obj=200,  # does not change if not None
        )
        mock_auto_create.assert_not_called()

        with self.assertRaises(ValueError) as error:
            batch.configure(
                batch_size=None,
                creation_time=-1,
                timeout_retries=10,
                dynamic=True,
            )
        check_error_message(self, error, value_error)
        self.check_instance(
            batch,
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            batching_type=None,
            recom_num_ref=200,  # does not change if not None
            recom_num_obj=200,  # does not change if not None
        )
        mock_auto_create.assert_not_called()

        #######################################################################
        # timeout_retries
        value_error = "'timeout_retries' must be positive, i.e. greater or equal that zero (>=0)."
        type_error = f"'timeout_retries' must be of type {int}."

        #######################################################################
        ## test wrong value
        with self.assertRaises(ValueError) as error:
            batch.configure(
                batch_size=None,
                creation_time=12.5,
                timeout_retries=-1,
                dynamic=True,
            )
        self.check_instance(
            batch,
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            batching_type=None,
            recom_num_ref=200,  # does not change if not None
            recom_num_obj=200,  # does not change if not None
        )
        check_error_message(self, error, value_error)

        #######################################################################
        ## test wrong type
        with self.assertRaises(TypeError) as error:
            batch.configure(
                batch_size=None,
                creation_time=12.5,
                timeout_retries=True,
                dynamic=True,
            )
        self.check_instance(
            batch,
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            batching_type=None,
            recom_num_ref=200,  # does not change if not None
            recom_num_obj=200,  # does not change if not None
        )
        check_error_message(self, error, type_error)

        with self.assertRaises(TypeError) as error:
            batch.configure(
                batch_size=None,
                creation_time=12.5,
                timeout_retries="12",
                dynamic=True,
            )
        self.check_instance(
            batch,
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            batching_type=None,
            recom_num_ref=200,  # does not change if not None
            recom_num_obj=200,  # does not change if not None
        )
        check_error_message(self, error, type_error)

        #######################################################################
        # dynamic
        type_error = "'dynamic' must be of type bool."

        with self.assertRaises(TypeError) as error:
            batch.configure(
                batch_size=100,
                creation_time=12.5,
                timeout_retries=10,
                dynamic=0,
            )
        check_error_message(self, error, type_error)
        self.check_instance(
            batch,
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            batching_type=None,
            recom_num_ref=200,  # does not change if not None
            recom_num_obj=200,  # does not change if not None
        )
        mock_auto_create.assert_not_called()

        #######################################################################
        # dynamic
        type_error = f"'batch_size' must be of type {int}."
        value_error = "'batch_size' must be positive, i.e. greater that zero (>0)."

        with self.assertRaises(TypeError) as error:
            batch.configure(
                batch_size=False,
                creation_time=12.5,
                timeout_retries=10,
                dynamic=True,
            )
        check_error_message(self, error, type_error)
        self.check_instance(
            batch,
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            batching_type=None,
            recom_num_ref=200,  # does not change if not None
            recom_num_obj=200,  # does not change if not None
        )
        mock_auto_create.assert_not_called()

        with self.assertRaises(TypeError) as error:
            batch.configure(
                batch_size=10.6,
                creation_time=12.5,
                timeout_retries=10,
                dynamic=True,
            )
        check_error_message(self, error, type_error)
        self.check_instance(
            batch,
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            batching_type=None,
            recom_num_ref=200,  # does not change if not None
            recom_num_obj=200,  # does not change if not None
        )
        mock_auto_create.assert_not_called()

        with self.assertRaises(ValueError) as error:
            batch.configure(
                batch_size=0,
                creation_time=12.5,
                timeout_retries=10,
                dynamic=True,
            )
        check_error_message(self, error, value_error)
        self.check_instance(
            batch,
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            batching_type=None,
            recom_num_ref=200,  # does not change if not None
            recom_num_obj=200,  # does not change if not None
        )
        mock_auto_create.assert_not_called()

        with self.assertRaises(ValueError) as error:
            batch.configure(
                batch_size=-10,
                creation_time=12.5,
                timeout_retries=10,
                dynamic=True,
            )
        check_error_message(self, error, value_error)
        self.check_instance(
            batch,
            batch_size=None,
            creation_time=12.5,
            timeout_retries=10,
            batching_type=None,
            recom_num_ref=200,  # does not change if not None
            recom_num_obj=200,  # does not change if not None
        )
        mock_auto_create.assert_not_called()

    def test_pop_methods(self):
        """
        Test the `pop_object` and the `pop_reference`.
        """

        batch = Batch(mock_connection_func())

        # mock BatchRequests
        mock_obj = Mock()
        mock_ref = Mock()
        batch._objects_batch = mock_obj
        batch._reference_batch = mock_ref

        mock_obj.pop.assert_not_called()
        mock_ref.pop.assert_not_called()

        # pop object default value
        batch.pop_object()
        mock_obj.pop.assert_called_with(-1)
        mock_ref.pop.assert_not_called()
        # reset mock objects
        mock_obj.reset_mock()
        mock_ref.reset_mock()

        # pop object at index
        batch.pop_object(10)
        mock_obj.pop.assert_called_with(10)
        mock_ref.pop.assert_not_called()
        # reset mock objects
        mock_obj.reset_mock()
        mock_ref.reset_mock()

        # pop reference default value
        batch.pop_reference()
        mock_obj.pop.assert_not_called()
        mock_ref.pop.assert_called_with(-1)
        # reset mock objects
        mock_obj.reset_mock()
        mock_ref.reset_mock()

        # pop reference at index
        batch.pop_reference(9)
        mock_obj.pop.assert_not_called()
        mock_ref.pop.assert_called_with(9)

    def test_empty_methods(self):
        """
        Test the `empty_objects` and the `empty_references`.
        """

        batch = Batch(mock_connection_func())

        # mock BatchRequests
        mock_obj = Mock()
        mock_ref = Mock()
        batch._objects_batch = mock_obj
        batch._reference_batch = mock_ref

        mock_obj.empty.assert_not_called()
        mock_ref.empty.assert_not_called()

        # empty objects
        batch.empty_objects()
        mock_obj.empty.assert_called()
        mock_ref.empty.assert_not_called()
        # reset mock objects
        mock_obj.reset_mock()
        mock_ref.reset_mock()

        # empty references
        batch.empty_references()
        mock_obj.empty.assert_not_called()
        mock_ref.empty.assert_called()

    def test_is_empty_methods(self):
        """
        Test the `is_empty_objects` and the `is_empty_references`.
        """

        batch = Batch(mock_connection_func())

        # mock BatchRequests
        mock_obj = Mock()
        mock_ref = Mock()
        batch._objects_batch = mock_obj
        batch._reference_batch = mock_ref

        mock_obj.is_empty.assert_not_called()
        mock_ref.is_empty.assert_not_called()

        # check if is_empty objects
        batch.is_empty_objects()
        mock_obj.is_empty.assert_called()
        mock_ref.is_empty.assert_not_called()
        # reset mock objects
        mock_obj.reset_mock()
        mock_ref.reset_mock()

        # check if is_empty reference
        batch.is_empty_references()
        mock_obj.is_empty.assert_not_called()
        mock_ref.is_empty.assert_called()



import json
import os
import uuid
from typing import Optional, List

import pytest
from pytest import FixtureRequest

import weaviate
from weaviate.data.replication import ConsistencyLevel

schema = {
    "classes": [
        {
            "class": "Ship",
            "description": "object",
            "properties": [
                {"dataType": ["string"], "description": "name", "name": "name"},
                {"dataType": ["string"], "description": "description", "name": "description"},
                {"dataType": ["int"], "description": "size", "name": "size"},
            ],
        }
    ]
}

SHIPS = [
    {
        "props": {"name": "HMS British Name", "size": 5, "description": "Super long description"},
        "id": uuid.uuid4(),
    },
    {
        "props": {
            "name": "The dragon ship",
            "size": 20,
            "description": "Interesting things about dragons",
        },
        "id": uuid.uuid4(),
    },
    {
        "props": {"name": "Blackbeard", "size": 43, "description": "Background info about movies"},
        "id": uuid.uuid4(),
    },
    {"props": {"name": "Titanic", "size": 1, "description": "Everyone knows"}, "id": uuid.uuid4()},
    {
        "props": {"name": "Artemis", "size": 34, "description": "Name from some story"},
        "id": uuid.uuid4(),
    },
    {
        "props": {
            "name": "The Crusty Crab",
            "size": 303,
            "description": "sponges, sponges, sponges",
        },
        "id": uuid.uuid4(),
    },
]


@pytest.fixture(scope="function")
def people_schema() -> str:
    with open(os.path.join(os.path.dirname(__file__), "people_schema.json"), encoding="utf-8") as f:
        return json.load(f)


@pytest.fixture(scope="module")
def client(request):
    port = 8080
    opts = parse_client_options(request)
    if opts:
        if opts.get("cluster"):
            port = 8087
            for _, c in enumerate(schema["classes"]):
                c["replicationConfig"] = {"factor": 2}

    client = weaviate.Client(f"http://localhost:{port}")
    client.schema.delete_all()
    client.schema.create(schema)
    with client.batch as batch:
        for ship in SHIPS:
            batch.add_data_object(ship["props"], "Ship", ship["id"])

        batch.flush()

    yield client
    client.schema.delete_all()


def parse_client_options(request: FixtureRequest) -> dict:
    try:
        if isinstance(request.param, dict):
            return request.param
    except AttributeError:
        return


def test_get_data(client):
    """Test GraphQL's Get clause."""
    where_filter = {"path": ["size"], "operator": "LessThan", "valueInt": 10}
    result = client.query.get("Ship", ["name", "size"]).with_limit(2).with_where(where_filter).do()
    objects = get_objects_from_result(result)
    a_found = False
    d_found = False
    for obj in objects:
        if obj["name"] == "HMS British Name":
            a_found = True
        if obj["name"] == "Titanic":
            d_found = True
    assert a_found and d_found and len(objects) == 2


def test_get_data_after(client):
    full_results = client.query.get("Ship", ["name"]).with_additional(["id"]).do()
    for i, ship in enumerate(full_results["data"]["Get"]["Ship"][:-1]):
        result = (
            client.query.get("Ship", ["name"])
            .with_additional(["id"])
            .with_limit(1)
            .with_after(ship["_additional"]["id"])
            .do()
        )
        assert (
            result["data"]["Get"]["Ship"][0]["_additional"]["id"]
            == full_results["data"]["Get"]["Ship"][i + 1]["_additional"]["id"]
        )


def test_get_data_after_wrong_types(client):
    with pytest.raises(TypeError):
        client.query.get("Ship", ["name"]).with_additional(["id"]).with_limit(1).with_after(
            1234
        ).do()


def test_multi_get_data(client, people_schema):
    """Test GraphQL's MultiGet clause."""
    client.schema.create(people_schema)
    client.data_object.create(
        {
            "name": "John",
        },
        "Person",
    )
    result = client.query.multi_get(
        [
            client.query.get("Ship", ["name"])
            .with_alias("one")
            .with_sort({"path": ["name"], "order": "asc"}),
            client.query.get("Ship", ["size"])
            .with_alias("two")
            .with_sort({"path": ["size"], "order": "asc"}),
            client.query.get("Person", ["name"]),
        ]
    ).do()["data"]["Get"]
    assert result["one"][0]["name"] == "Artemis"
    assert result["two"][0]["size"] == 1
    assert result["Person"][0]["name"] == "John"


def test_aggregate_data(client):
    """Test GraphQL's Aggregate clause."""
    where_filter = {"path": ["name"], "operator": "Equal", "valueString": "The dragon ship"}

    result = (
        client.query.aggregate("Ship")
        .with_where(where_filter)
        .with_group_by_filter(["name"])
        .with_fields("groupedBy {value}")
        .with_fields("name{count}")
        .do()
    )

    aggregation = get_aggregation_from_aggregate_result(result)
    assert "groupedBy" in aggregation, "Missing groupedBy"
    assert "name" in aggregation, "Missing name property"


def get_objects_from_result(result):
    return result["data"]["Get"]["Ship"]


def get_aggregation_from_aggregate_result(result):
    return result["data"]["Aggregate"]["Ship"][0]


@pytest.mark.parametrize("query", ["sponges", "sponges\n"])
def test_bm25(client, query):
    result = client.query.get("Ship", ["name"]).with_bm25(query, ["name", "description"]).do()
    assert len(result["data"]["Get"]["Ship"]) == 1
    assert result["data"]["Get"]["Ship"][0]["name"] == "The Crusty Crab"


def test_bm25_no_result(client):
    result = client.query.get("Ship", ["name"]).with_bm25("sponges\n", ["name"]).do()
    assert len(result["data"]["Get"]["Ship"]) == 0


@pytest.mark.parametrize("query", ["sponges", "sponges\n"])
def test_hybrid(client, query: str):
    """Test hybrid search with alpha=0.5 to have a combination of BM25 and vector search."""
    result = client.query.get("Ship", ["name"]).with_hybrid(query, alpha=0.5, vector=[1] * 300).do()

    # will find more results. "The Crusty Crab" is still first, because it matches with the BM25 search
    assert len(result["data"]["Get"]["Ship"]) >= 1
    assert result["data"]["Get"]["Ship"][0]["name"] == "The Crusty Crab"


@pytest.mark.parametrize(
    "properties,num_results",
    [(None, 1), ([], 1), (["description"], 1), (["description", "name"], 1), (["name"], 0)],
)
def test_hybrid_properties(client, properties: Optional[List[str]], num_results: int):
    """Test hybrid search with alpha=0.5 to have a combination of BM25 and vector search."""
    result = (
        client.query.get("Ship", ["name"])
        .with_hybrid("sponges", alpha=0.0, properties=properties)
        .do()
    )

    # "The Crusty Crab" has "sponges" in its description, it cannot be found in other properties
    if num_results > 0:
        assert len(result["data"]["Get"]["Ship"]) >= 1

        assert result["data"]["Get"]["Ship"][0]["name"] == "The Crusty Crab"
    else:
        assert len(result["data"]["Get"]["Ship"]) == 0


def test_group_by(client, people_schema):
    """Test hybrid search with alpha=0.5 to have a combination of BM25 and vector search."""
    client.schema.delete_all()
    client.schema.create(people_schema)

    persons = []
    for i in range(10):
        persons.append(uuid.uuid4())
        client.data_object.create({"name": "randomName" + str(i)}, "Person", persons[-1])

    client.data_object.create({}, "Call", "3ab05e06-2bb2-41d1-b5c5-e044f3aa9623")
    client.data_object.create({}, "Call", "3ab05e06-2bb2-41d1-b5c5-e044f3aa9622")

    # create refs
    for i in range(5):
        client.data_object.reference.add(
            to_uuid=persons[i],
            from_property_name="caller",
            from_uuid="3ab05e06-2bb2-41d1-b5c5-e044f3aa9623"
            if i % 2 == 0
            else "3ab05e06-2bb2-41d1-b5c5-e044f3aa9622",
            from_class_name="Call",
            to_class_name="Person",
        )

    result = (
        client.query.get("Call", ["caller{... on Person{name}}"])
        .with_near_object({"id": "3ab05e06-2bb2-41d1-b5c5-e044f3aa9622"})
        .with_group_by(properties=["caller"], groups=2, objects_per_group=3)
        .with_additional("group{hits {_additional{vector}caller{... on Person{name}}}}")
        .do()
    )

    # will find more results. "The Crusty Crab" is still first, because it matches with the BM25 search
    assert len(result["data"]["Get"]["Call"]) >= 1
    assert result["data"]["Get"]["Call"][0]["caller"][0]["name"] == "randomName0"


@pytest.mark.parametrize(
    "client,level",
    [
        ({"cluster": True}, ConsistencyLevel.ONE),
        ({"cluster": True}, ConsistencyLevel.QUORUM),
        ({"cluster": True}, ConsistencyLevel.ALL),
    ],
    indirect=["client"],
)
def test_consistency_level(client, level):
    result = (
        client.query.get("Ship", ["name"])
        .with_consistency_level(level)
        .with_additional("isConsistent")
        .do()
    )
    for _, res in enumerate(get_objects_from_result(result)):
        assert res["_additional"]["isConsistent"]


@pytest.mark.parametrize(
    "single,grouped",
    [
        ("Describe the following as a Facebook Ad: {review}", None),
        (None, "Describe the following as a LinkedIn Ad: {review}"),
        (
            "Describe the following as a Twitter Ad: {review}",
            "Describe the following as a Mastodon Ad: {review}",
        ),
        (
            "Describe the following as a Twitter Ad: \n Review: {review} \n Name: {name}",
            "Describe the following as a Mastodon Ad:  \n Review: {review} \n Name: {name}",
        ),
    ],
)
def test_generative_openai(single: str, grouped: str):
    """Test client credential flow with various providers."""
    api_key = os.environ.get("OPENAI_APIKEY")
    if api_key is None:
        pytest.skip("No OpenAI API key found.")

    client = weaviate.Client(
        "http://127.0.0.1:8086", additional_headers={"X-OpenAI-Api-Key": api_key}
    )
    client.schema.delete_all()
    wine_class = {
        "class": "Wine",
        "properties": [
            {"name": "name", "dataType": ["string"]},
            {"name": "review", "dataType": ["string"]},
        ],
        "moduleConfig": {"generative-openai": {}},
    }
    client.schema.create_class(wine_class)
    client.data_object.create(
        data_object={"name": "Super expensive wine", "review": "Tastes like a fresh ocean breeze"},
        class_name="Wine",
    )
    client.data_object.create(
        data_object={"name": "cheap wine", "review": "Tastes like forest"}, class_name="Wine"
    )

    result = (
        client.query.get("Wine", ["name", "review"])
        .with_generate(single_prompt=single, grouped_task=grouped)
        .do()
    )
    assert result["data"]["Get"]["Wine"][0]["_additional"]["generate"]["error"] is None

    grouped_properties = ["review"]
    result = (
        client.query.get("Wine", ["name", "review"])
        .with_generate(
            single_prompt=single, grouped_task=grouped, grouped_properties=grouped_properties
        )
        .do()
    )
    assert result["data"]["Get"]["Wine"][0]["_additional"]["generate"]["error"] is None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: weaviate.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0eweaviate.proto\x12\x0cweaviategrpc\x1a\x1cgoogle/protobuf/struct.proto\"\xfb\x02\n\rSearchRequest\x12\x12\n\nclass_name\x18\x01 \x01(\t\x12\r\n\x05limit\x18\x02 \x01(\r\x12\x41\n\x15\x61\x64\x64itional_properties\x18\x03 \x01(\x0b\x32\".weaviategrpc.AdditionalProperties\x12\x33\n\x0bnear_vector\x18\x04 \x01(\x0b\x32\x1e.weaviategrpc.NearVectorParams\x12\x33\n\x0bnear_object\x18\x05 \x01(\x0b\x32\x1e.weaviategrpc.NearObjectParams\x12,\n\nproperties\x18\x06 \x01(\x0b\x32\x18.weaviategrpc.Properties\x12\x37\n\rhybrid_search\x18\x07 \x01(\x0b\x32 .weaviategrpc.HybridSearchParams\x12\x33\n\x0b\x62m25_search\x18\x08 \x01(\x0b\x32\x1e.weaviategrpc.BM25SearchParams\"\xb4\x01\n\x14\x41\x64\x64itionalProperties\x12\x0c\n\x04uuid\x18\x01 \x01(\x08\x12\x0e\n\x06vector\x18\x02 \x01(\x08\x12\x18\n\x10\x63reationTimeUnix\x18\x03 \x01(\x08\x12\x1a\n\x12lastUpdateTimeUnix\x18\x04 \x01(\x08\x12\x10\n\x08\x64istance\x18\x05 \x01(\x08\x12\x11\n\tcertainty\x18\x06 \x01(\x08\x12\r\n\x05score\x18\x07 \x01(\x08\x12\x14\n\x0c\x65xplainScore\x18\x08 \x01(\x08\"]\n\nProperties\x12\x1a\n\x12non_ref_properties\x18\x01 \x03(\t\x12\x33\n\x0eref_properties\x18\x02 \x03(\x0b\x32\x1b.weaviategrpc.RefProperties\"V\n\x12HybridSearchParams\x12\r\n\x05query\x18\x01 \x01(\t\x12\x12\n\nproperties\x18\x02 \x03(\t\x12\x0e\n\x06vector\x18\x03 \x03(\x02\x12\r\n\x05\x61lpha\x18\x04 \x01(\x02\"5\n\x10\x42M25SearchParams\x12\r\n\x05query\x18\x01 \x01(\t\x12\x12\n\nproperties\x18\x02 \x03(\t\"v\n\rRefProperties\x12\x14\n\x0clinked_class\x18\x01 \x01(\t\x12\x1a\n\x12reference_property\x18\x02 \x01(\t\x12\x33\n\x11linked_properties\x18\x03 \x01(\x0b\x32\x18.weaviategrpc.Properties\"l\n\x10NearVectorParams\x12\x0e\n\x06vector\x18\x01 \x03(\x02\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x42\x0c\n\n_certaintyB\x0b\n\t_distance\"h\n\x10NearObjectParams\x12\n\n\x02id\x18\x01 \x01(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x42\x0c\n\n_certaintyB\x0b\n\t_distance\"H\n\x0bSearchReply\x12+\n\x07results\x18\x01 \x03(\x0b\x32\x1a.weaviategrpc.SearchResult\x12\x0c\n\x04took\x18\x02 \x01(\x02\"\x86\x01\n\x0cSearchResult\x12\x32\n\nproperties\x18\x01 \x01(\x0b\x32\x1e.weaviategrpc.ResultProperties\x12\x42\n\x15\x61\x64\x64itional_properties\x18\x02 \x01(\x0b\x32#.weaviategrpc.ResultAdditionalProps\"\xef\x02\n\x15ResultAdditionalProps\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x06vector\x18\x02 \x03(\x02\x12\x1a\n\x12\x63reation_time_unix\x18\x03 \x01(\x04\x12\"\n\x1a\x63reation_time_unix_present\x18\x04 \x01(\x08\x12\x1d\n\x15last_update_time_unix\x18\x05 \x01(\x04\x12%\n\x1dlast_update_time_unix_present\x18\x06 \x01(\x08\x12\x10\n\x08\x64istance\x18\x07 \x01(\x02\x12\x18\n\x10\x64istance_present\x18\x08 \x01(\x08\x12\x11\n\tcertainty\x18\t \x01(\x02\x12\x19\n\x11\x63\x65rtainty_present\x18\n \x01(\x08\x12\r\n\x05score\x18\x0b \x01(\x02\x12\x15\n\rscore_present\x18\x0c \x01(\x08\x12\x15\n\rexplain_score\x18\r \x01(\t\x12\x1d\n\x15\x65xplain_score_present\x18\x0e \x01(\x08\"\x91\x01\n\x10ResultProperties\x12\x33\n\x12non_ref_properties\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x34\n\tref_props\x18\x02 \x03(\x0b\x32!.weaviategrpc.ReturnRefProperties\x12\x12\n\nclass_name\x18\x03 \x01(\t\"\\\n\x13ReturnRefProperties\x12\x32\n\nproperties\x18\x01 \x03(\x0b\x32\x1e.weaviategrpc.ResultProperties\x12\x11\n\tprop_name\x18\x02 \x01(\t2N\n\x08Weaviate\x12\x42\n\x06Search\x12\x1b.weaviategrpc.SearchRequest\x1a\x19.weaviategrpc.SearchReply\"\x00\x42#Z!github.com/weaviate/weaviate/grpcb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'weaviate_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z!github.com/weaviate/weaviate/grpc'
  _SEARCHREQUEST._serialized_start=63
  _SEARCHREQUEST._serialized_end=442
  _ADDITIONALPROPERTIES._serialized_start=445
  _ADDITIONALPROPERTIES._serialized_end=625
  _PROPERTIES._serialized_start=627
  _PROPERTIES._serialized_end=720
  _HYBRIDSEARCHPARAMS._serialized_start=722
  _HYBRIDSEARCHPARAMS._serialized_end=808
  _BM25SEARCHPARAMS._serialized_start=810
  _BM25SEARCHPARAMS._serialized_end=863
  _REFPROPERTIES._serialized_start=865
  _REFPROPERTIES._serialized_end=983
  _NEARVECTORPARAMS._serialized_start=985
  _NEARVECTORPARAMS._serialized_end=1093
  _NEAROBJECTPARAMS._serialized_start=1095
  _NEAROBJECTPARAMS._serialized_end=1199
  _SEARCHREPLY._serialized_start=1201
  _SEARCHREPLY._serialized_end=1273
  _SEARCHRESULT._serialized_start=1276
  _SEARCHRESULT._serialized_end=1410
  _RESULTADDITIONALPROPS._serialized_start=1413
  _RESULTADDITIONALPROPS._serialized_end=1780
  _RESULTPROPERTIES._serialized_start=1783
  _RESULTPROPERTIES._serialized_end=1928
  _RETURNREFPROPERTIES._serialized_start=1930
  _RETURNREFPROPERTIES._serialized_end=2022
  _WEAVIATE._serialized_start=2024
  _WEAVIATE._serialized_end=2102
# @@protoc_insertion_point(module_scope)



import unittest

from requests.exceptions import ConnectionError as RequestsConnectionError

from test.util import mock_connection_func, check_error_message, check_startswith_error_message
from weaviate.exceptions import UnexpectedStatusCodeException
from weaviate.gql.aggregate import AggregateBuilder


class TestAggregateBuilder(unittest.TestCase):
    def setUp(self):

        self.aggregate = AggregateBuilder("Object", None)

    def test_with_meta_count(self):
        """
        Test the `with_meta_count` method.
        """

        query = self.aggregate.with_meta_count().build()
        self.assertEqual("{Aggregate{Object{meta{count}}}}", query)

    def test_with_fields(self):
        """
        Test the `with_fields` method.
        """

        query = self.aggregate.with_fields("size { mean }").build()
        self.assertEqual("{Aggregate{Object{size { mean }}}}", query)

    def test_with_where(self):
        """
        Test the `with_where` method.
        """

        query = (
            self.aggregate.with_meta_count()
            .with_where({"operator": "Equal", "valueString": "B", "path": ["name"]})
            .build()
        )
        self.assertEqual(
            '{Aggregate{Object(where: {path: ["name"] operator: Equal valueString: "B"} ){meta{count}}}}',
            query,
        )

    def test_group_by_filter(self):
        """
        Test the `with_group_by_filter` method.
        """

        query = (
            self.aggregate.with_group_by_filter(["name"])
            .with_fields("groupedBy { value }")
            .with_fields("name { count }")
            .build()
        )
        self.assertEqual(
            '{Aggregate{Object(groupBy: ["name"]){groupedBy { value }name { count }}}}', query
        )

    def test_do(self):
        """
        Test the `do` method.
        """

        # test exceptions
        requests_error_message = "Query was not successful."

        # requests.exceptions.ConnectionError
        mock_obj = mock_connection_func("post", side_effect=RequestsConnectionError("Test"))
        self.aggregate._connection = mock_obj
        with self.assertRaises(RequestsConnectionError) as error:
            self.aggregate.do()
        check_error_message(self, error, requests_error_message)

        # weaviate.UnexpectedStatusCodeException
        mock_obj = mock_connection_func("post", status_code=204)
        self.aggregate._connection = mock_obj
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            self.aggregate.do()
        check_startswith_error_message(self, error, "Query was not successful")

        filter_name = {"path": ["name"], "operator": "Equal", "valueString": "B"}

        self.aggregate.with_group_by_filter(["name"]).with_fields(
            "groupedBy { value }"
        ).with_fields("name { count }").with_where(filter_name)
        expected_gql_clause = '{Aggregate{Object(where: {path: ["name"] operator: Equal valueString: "B"} groupBy: ["name"]){groupedBy { value }name { count }}}}'

        mock_obj = mock_connection_func("post", status_code=200, return_json={"status": "OK!"})
        self.aggregate._connection = mock_obj
        self.assertEqual(self.aggregate.do(), {"status": "OK!"})
        mock_obj.post.assert_called_with(
            path="/graphql", weaviate_object={"query": expected_gql_clause}
        )

    def test_uncapitalized_class_name(self):
        """
        Test the uncapitalized class_name.
        """

        aggregate = AggregateBuilder("Test", None)
        self.assertEqual(aggregate._class_name, "Test")

        aggregate = AggregateBuilder("test", None)
        self.assertEqual(aggregate._class_name, "Test")



import pytest

import weaviate

GIT_HASH = "c804a73"
SERVER_VERSION = "1.19.3"
NODE_NAME = "node1"
NUM_OBJECT = 10

schema = {
    "classes": [
        {
            "class": "ClassA",
            "properties": [
                {"dataType": ["string"], "name": "stringProp"},
                {"dataType": ["int"], "name": "intProp"},
            ],
        }
    ]
}


@pytest.fixture(scope="module")
def client():
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    yield client
    client.schema.delete_all()


def test_get_nodes_status_without_data(client):
    """get nodes status without data"""
    resp = client.cluster.get_nodes_status()
    assert len(resp) == 1
    assert resp[0]["gitHash"] == GIT_HASH
    assert resp[0]["name"] == NODE_NAME
    assert len(resp[0]["shards"]) == 0
    assert resp[0]["stats"]["objectCount"] == 0
    assert resp[0]["stats"]["shardCount"] == 0
    assert resp[0]["status"] == "HEALTHY"
    assert resp[0]["version"] == SERVER_VERSION


def test_get_nodes_status_with_data(client):
    """get nodes status with data"""
    class_name = "ClassA"
    client.schema.create(schema)
    for i in range(NUM_OBJECT):
        client.data_object.create({"stringProp": f"object-{i}", "intProp": i}, "ClassA")

    resp = client.cluster.get_nodes_status()
    assert len(resp) == 1
    assert resp[0]["gitHash"] == GIT_HASH
    assert resp[0]["name"] == NODE_NAME
    assert len(resp[0]["shards"]) == 1
    assert resp[0]["shards"][0]["class"] == class_name
    assert resp[0]["shards"][0]["objectCount"] == NUM_OBJECT
    assert resp[0]["stats"]["objectCount"] == NUM_OBJECT
    assert resp[0]["stats"]["shardCount"] == 1
    assert resp[0]["status"] == "HEALTHY"
    assert resp[0]["version"] == SERVER_VERSION
